## May 29. Changes from version 10: removed zero-padding for pitchAutocor as super hard to fix (maybe it's possible, I don't know, but I give up!). Added an additional check in pitchCep (>1.8 times absolute peak, to avoid false subharmonics) and adaptive width of interval for finding peaks in cepstrum.
# To do: Combine with syllable segmenting function, plot osc under spectrograms with syllables shown

library(tuneR)
library(plyr)
library(zoo)
library(seewave) 

#
# *******************************
# H E L P E R   F U N C T I O N S (for user-called functions, see lines 500+)
# *******************************
# 

isCentral.localMax = function(x, threshold){
  # internal helper function, needed for peak detection with rollapply from zoo package (less versatile but x10 faster than fpeaks from seewave)
  middle = ceiling(length(x)/2)
  if (which.max(x)==middle & x[middle]>threshold){
    return (TRUE)
  } else {
    return (FALSE)
  }
}

costJumps = function(cand1, cand2){
  # internal helper function for calculating the cost of transitions between pitch candidates. Needed for postprocessing - finding the optimal pitch contour
  return ( 1/(1+10*exp(3-7*abs(cand1-cand2))) )
  # visualization: a = seq(-3,3,by=.01); b = 1/(1+10*exp(3-7*abs(a))); plot (a,b,type='l')
}

findGrad = function (path, interpol=3){ 
  # internal helper function for postprocessing. Returns gradient of internal energy of a snake (see below)
  
  # interpolate 2 values before the first one and two after the last one based on /interpol/ number of points
  interpol = ifelse (interpol>length(path), length(path), interpol) # in case the path is shorter than the specified interpol
  if (interpol==1){
    path = c(rep(path[1],2), path, rep(path[length(path)],2))
  } else {
    slopeLeft = summary(lm(path[1:interpol]~seq(1,interpol)))$coef[2,1]
    minus12 = path[1] - c(1,2)*slopeLeft
    slopeRight = summary(lm( path[(length(path)-interpol+1) : length(path)]~seq(1,interpol)))$coef[2,1]
    plus12 = path[length(path)] + c(1,2)*slopeRight
    path = c (minus12[2], minus12[1], path, plus12[1], plus12[2])
  }
  
  # take 4th derivative of the path with interpolated values (so that we get d4f over the entire length of the original path)
  grad = rep(0, length(path))
  for (i in 3:(length(path)-2)) {
    grad[i] = path[i-2]-4*path[i-1]+6*path[i]-4*path[i+1]+path[i+2]
  }
  grad = grad[3: (length(grad)-2)]
  return (grad)
}


forcePerPath = function (pitch, pitchCands, pitchCert, pitchCenterGravity, certWeight){
  # internal helper function for postprocessing. Returns the total force acting on a snake (sum of internal and external gradients, ie of the elastic force trying to straighten the snake [internal] and of the force pushing the snake towards the most certain pitch estimates [external])
  ran = diff(range(pitchCands, na.rm=T))
  # pitchForce = -(pitch_path - pitchCenterGravity) / ran
  pitchForce = pitch # just a quick way to initialize a vector of the right length
  for (i in 1:ncol(pitchCands)){
    cands = na.omit(pitchCands[,i])
    certs = na.omit(pitchCert[,i])
    deltas = 1/exp((cands-pitch[i])^2)
    forces = certs * deltas
    forces = ifelse (cands>pitch[i], forces, -forces)
    pitchForce[i] = sum(forces)
  }
  internalForce = -findGrad(pitch)
  return ( certWeight*pitchForce + (1-certWeight)*internalForce )
}

snake = function (pitch, pitchCands, pitchCert, certWeight, snakeSmoothingStep=0.05, snakeIterMultiplier=2, plotSnake=F) {
  # internal helper function for postprocessing. Wriggles a snake along the gradient of internal+external forces. NB: if the snake is run, the final contour may deviate from the actually measured pitch candidates!
  
  ran = diff(range(pitchCands, na.rm=T)) # range of pitch
  maxIter = ran / snakeSmoothingStep * snakeIterMultiplier
  
  # plot for debugging or esthetic appreciation
  if (plotSnake){
    plot(seq(1,ncol(pitchCands)),pitch, type='n', ylim=c( range(pitchCands, na.rm=T)[1]-.3*ran, range(pitchCands, na.rm=T)[2]+.3*ran))
    for (r in 1:nrow(pitchCands)){
      points (seq(1,ncol(pitchCands)), pitchCands[r,], cex=as.numeric(pitchCert[r,])*2)
    }
    lines (seq(1,ncol(pitchCands)), pitch)
  }
  
  # optimization algorithm follows
  for (i in 1:maxIter){
    force = forcePerPath (pitch, pitchCands, pitchCert, pitchCenterGravity, certWeight)
    pitch = pitch + snakeSmoothingStep*force
    if (plotSnake) {lines(seq(1,length(pitch)),pitch, type='l', col='green', lty=4)}
  }
  
  if (plotSnake){
    lines(seq(1,length(pitch)),pitch, type='l', col='blue', lwd=3)
  }
  return (pitch)
}


pathfinder = function(pitchCands, pitchCert, certWeight=0.5, interpolWindow=3, interpolTolerance=0.05, interpolCert=0.3, runSnake=T, snakeSmoothingStep=0.05, snakeIterMultiplier=2, plotSnake=F){
  # internal helper function for postprocessing. Starts with a reasonable guess and computes the ~optimal (not quite - too computationally expensive) pitch contour.
  
  # take log to approximate human perception of pitch differences
  pitchCands[!is.na(pitchCands)] = log2(pitchCands[!is.na(pitchCands)])
  
  # Interpolation: if a frame has no pitch candidate at all (NA) or no candidate between the most likely candidates for the adjacent frames, add such a candidate with ~low certainty
  for (f in 1:ncol(pitchCands)){  
    pitchCenterGravity = apply (as.matrix(1:ncol(pitchCands),nrow=1), 1, function(x)mean(pitchCands[,x], weights=pitchCert[,x]/sum(pitchCert[,x]), na.rm=T)) # NB: in the loop b/c it has to be done recursively, taking interpolated values into account for further interpolation. Unfortunate but can't be helped
    left = max(1, f-interpolWindow); right = min(ncol(pitchCands),f+interpolWindow)
    med = median(pitchCenterGravity[left:right], na.rm=T) # median over interpolation window (by default ±2 points)
    if (sum(pitchCands[,f]>(1-interpolTolerance)*med & pitchCands[,f]<(1+interpolTolerance)*med, na.rm=T)==0){ # defaults to (0.95*med, 1.05*med)
      pitchCands = rbind(pitchCands, rep(NA, ncol(pitchCands)))
      pitchCert = rbind (pitchCert, rep(NA, ncol(pitchCert)))
      pitchCands[nrow(pitchCands),f] = median(pitchCenterGravity[left:right], na.rm=T)
      pitchCert[nrow(pitchCert),f] = interpolCert # certainty assigned to interpolated frames
    }
  }  
  
  # order pitch candidates and certainties in each frame from lowest to highest pitch (helpful for further processing)
  o = apply (as.matrix(1:ncol(pitchCands), nrow=1), 1, function(x) order(pitchCands[,x]))
  pitchCands = apply (as.matrix(1:ncol(pitchCands), nrow=1), 1, function(x) pitchCands[o[,x],x])
  pitchCert = apply (as.matrix(1:ncol(pitchCert), nrow=1), 1, function(x) pitchCert[o[,x],x])
  pitchCands = pitchCands[rowSums(!is.na(pitchCands))!=0, ] # remove rows with all NA's
  pitchCert = pitchCert[rowSums(!is.na(pitchCert))!=0, ]
  
  # special case: only a single pitch candidate for all frames in a syllable (no paths to chose among)
  if (class(pitchCands)=='numeric'){return(2^pitchCands)}
  
  # get the "center of gravity" of pitch candidates in each frame (mean of all pitch candidates weighted by their respective certainties)
  pitchCenterGravity = apply (as.matrix(1:ncol(pitchCands),nrow=1), 1, function(x)mean(pitchCands[,x], weights=pitchCert[,x]/sum(pitchCert[,x]), na.rm=T))
  
  # start at the beginning of the snake: find the most plausible starting pitch by taking median over the first few frames, weighted by certainty
  p = median ( pitchCenterGravity[1:min(5,length(pitchCenterGravity))] )
  c = pitchCert[,1]/abs(pitchCands[,1]-p) # b/c there may be NA's, and they can't be excluded directly in which.max in the next line
  point_current = pitchCands[which.max (c),1]
  path = point_current
  costPathForward = 0
  
  for (i in 2:ncol(pitchCands)){
    cands = na.omit(pitchCands[,i])
    cost_cert = abs(cands-pitchCenterGravity[i])
    cost_pitchJump = apply ( as.matrix(1:length(cands),nrow=1), 1, function(x)costJumps(point_current, cands[x]) )
    costs = certWeight*cost_cert+(1-certWeight)*cost_pitchJump
    path = c(path, pitchCands[which.min(costs),i])
    costPathForward = costPathForward + min(costs)
  }
  
  # run backwards
  pitchCands_rev = pitchCands[,rev(1:ncol(pitchCands))]
  pitchCert_rev = pitchCert[,rev(1:ncol(pitchCert))]
  pitchCenterGravity_rev = rev(pitchCenterGravity)
  
  p = median ( pitchCenterGravity_rev[1:min(5,length(pitchCenterGravity_rev))] )
  c = na.omit (pitchCert_rev[,1]/abs(pitchCands_rev[,1]-p)) # b/c there may be NA's, and they can't be excluded directly in which.max in the next line
  point_current = pitchCands_rev[which.max (c),1]
  path_rev = point_current
  costPathBackward = 0
  
  for (i in 2:ncol(pitchCands_rev)){
    cands = na.omit(pitchCands_rev[,i])
    cost_cert = abs(cands-pitchCenterGravity_rev[i])
    cost_pitchJump = apply ( as.matrix(1:length(cands),nrow=1), 1, function(x)costJumps(point_current, cands[x]) )
    costs = certWeight*cost_cert+(1-certWeight)*cost_pitchJump
    path_rev = c(path_rev, pitchCands_rev[which.min(costs),i])
    costPathBackward = costPathBackward + min(costs)
  }
  
  if (costPathForward<costPathBackward){
    pitch_final = path
  } else {
    pitch_final = rev(path_rev)
  }
  
  if (runSnake){
    pitch_final = snake (pitch_final, pitchCands, pitchCert, certWeight, snakeSmoothingStep, snakeIterMultiplier, plotSnake)
  }
  
  return(2^pitch_final)
}


findVoicedSegments = function(pitchCands, shortestSyllable = 20, shortestBreak=50, step, samplingRate, pitchCandsPerFrameToQualifyAsPutativelyVoiced=2){ 
  # internal helper function for postprocessing. Merges voiced segments at least /shortestSyllable/ ms long and separated by less than /shortestBreak/ ms
  
  putativelyVoiced = apply (pitchCands, 2, function(x)ifelse(sum(!is.na(x)) >= pitchCandsPerFrameToQualifyAsPutativelyVoiced, 1, NA)) # 
  noRequired = max (2, ceiling ( shortestSyllable / step)) # the smallest number of consecutive non-NA pitch values that constitute a voiced segment; but at least 2
  noToleratedNA = floor ( shortestBreak / step ) # the greatest number of NA values that we tolerate before we say a new voiced syllable begins
  
  # find and save separately all voiced segments 
  segmentStart = numeric(); segmentEnd = numeric()
  i = 1
  while (i < (length(putativelyVoiced)-noRequired+1)){
    # find beginning
    while (i < (length(putativelyVoiced)-noRequired+1) ){
      if (sum(!is.na(putativelyVoiced[i : (i+noRequired-1)])) == noRequired){
        segmentStart = c(segmentStart, i )
        break
      }
      i = i+1
    }
    # find end
    if (length(segmentEnd)<length(segmentStart)) {
      while (i < (length(putativelyVoiced)-noToleratedNA+1) ){
        if (sum(putativelyVoiced[i : (i+noToleratedNA)], na.rm=T) == 0){
          segmentEnd = c(segmentEnd, i-1)
          i = i-1
          break
        }
        i = i+1
      }
      if (length(segmentEnd)<length(segmentStart)) { # if the end is not found, take the last non-NA value
        segmentEnd = c ( segmentEnd, max(which(!is.na(putativelyVoiced))) )
        break
      }
    }
    i = i+1
  } 
  return (data.frame(segmentStart=segmentStart, segmentEnd=segmentEnd) )
}

to_dB = function(x){
  return (10*log10(x/(1-x)))
}

ftwindow_modif = function (wl, wn = "gaussian") 
  # as in seewave but with gaussian window option
{
  if (wn == "bartlett") 
    w <- bartlett.w(wl)
  if (wn == "blackman") 
    w <- blackman.w(wl)
  if (wn == "flattop") 
    w <- flattop.w(wl)
  if (wn == "hamming") 
    w <- hamming.w(wl)
  if (wn == "hanning") 
    w <- hanning.w(wl)
  if (wn == "rectangle") 
    w <- rectangle.w(wl)
  if (wn == "gaussian") 
    w <- gaussian.w(wl)
  return(w)
}

gaussian.w = function (n) 
{
  if (n <= 0) 
    stop("'n' must be a positive integer")
  s = 1:n
  w <- (exp(-12*(((1:n)/n)-0.5)^2)-exp(-12))/(1-exp(-12)) # Boersma (PRAAT)
  return(w)
}

getFrameBank = function(sound, samplingRate, windowLength_points, wn, step, zp, filter=NULL){
  # internal helper function. Saves windowed (and optionally zero-padded) frames, i.e. chunks of the sound file of the right size and spacing. Handy for further processing
  
  # normalize
  if (min(sound)>0) {sound = scale(sound)}
  sound = sound / max( abs(max(sound)), abs(min(sound))) # normalize to range from no less than -1 to no more than +1
  duration = length(sound) / samplingRate
  myseq = seq(1, length(sound) - windowLength_points, step/1000*samplingRate)
  if (is.null(filter)) {
    filter=ftwindow_modif(wl=windowLength_points, wn = wn)
  }
  zpExtra = floor( (zp-windowLength_points)/2) * 2 # how many extra zeroes we pad with. Made even       
  if (zpExtra>0){
    frameBank = apply (as.matrix(myseq), 1, function(x) c(rep(0,zpExtra/2), sound[x:(windowLength_points + x - 1)]*filter, rep(0,zpExtra/2)) )
  } else {
    frameBank = apply (as.matrix(myseq), 1, function(x) sound[x:(windowLength_points + x - 1)]*filter )
  }
  return (frameBank)
}

Mode <- function(x) { 
  # internal helper function for spectral (~BaNa) pitch tracker. NOT quite the same as simply mode(x)
  x = sort(x)
  ux <- unique(x)
  if (length(ux)<length(x)) { 
    return (ux[which.max(tabulate(match(x, ux)))])
  } else { # if every element is unique, return the smallest
    return (x[1])  
  }
}

getEntropy = function(x) { # returns entropy of a vector on a scale 0 to 1
  x = ifelse (x==0, 1e-10, x)  # otherwise log0 gives NaN
  return ( -sum(log2(x/sum(x))*x/sum(x)) / log(length(x))*log(2) )
}

medianSmoother = function (df, smoothing_ww, smoothing_threshold) {
  # internal helper function for smoothing pitch contours or other contours. Expects a matrix or something convertible to a matrix
  # smoothing_ww specifies the number of points that are averaged
  # smoothing_threshold specifies the tolerated deviance, in semitones
  
  df = as.matrix(df)
  temp = df # to calculate median_over_window for original values
  hw = floor(smoothing_ww/2) # smooth over ± half the smoothing_ww
  for (i in 1:nrow(df)) {
    window = c ( max(i-hw, 1), min(i+hw,nrow(df))) # smoothing window
    median_over_window = apply (as.matrix(temp[ (window[1]:window[2]),]), 2, function(x) median(x, na.rm = T) ) # NB: use either temp or df, for original or smoothed values to be used for calculating median_over_window    
    deviance = 12*log2(df[i,] / median_over_window) # difference from median pitch etc over window, in semitones
    cond = abs(deviance-1) > smoothing_threshold
    if (mean(!is.na(cond))>0) { # if we need to smooth at least 1 var in this frame
      for (c in 1:ncol(df)){
        df[i,c] = ifelse (cond[c]==T, median_over_window[c], df[i,c] )
      }
    }
  }
  return (df)
}

ratios = data.frame (name = c('F1/F0','F2/FO','F2/F1','F3/F0','F3/F2','F4/F0','F4/F1','F4/F2','F4/F3'), value_low = c(1.9, 2.8, 1.42, 3.8, 1.29, 4.8, 2.4, 1.59, 1.15), value_high = c(2.1, 3.2, 1.59, 4.2, 1.42, 5.2, 2.6, 1.8, 1.29), divide_lower_by = c(1, 1, 2, 1, 3, 1, 2, 3, 4)) # empirically observed ratios of harmonics, from BaNa algorithm: Ba, H., Yang, N., Demirkol, I., & Heinzelman, W. (2012, August). BaNa: A hybrid approach for noise resilient pitch detection. In Statistical Signal Processing Workshop (SSP), 2012 IEEE (pp. 369-372). IEEE.

analyzeFrame = function (frame, autoCorrelation=NULL, samplingRate=44100, windowLength_points=2048, wn='gaussian', zp=0, zpCep=2^13, cutoff_dom=6000, voiced_threshold_autocor=0.75, voiced_threshold_cep=0.45, voiced_threshold_spec=0.45, specPitchThreshold_nullNA=0.8, slopeSpec=0.75, widthSpec=100, pitch_floor=75, pitch_ceiling=3500, silence=0.03, dom_threshold=0.1, pitchSpec_only_peak_weight=0.6, autoCorrelation_filter=NULL, entropyThreshold=0.9, maxNoCands=1) {
  ## internal helper function. Analyzes a frame and returns several measures of pitch plus other acoustic features.
  # frame: a simple vector of normalized amplitudes: readWave(soundfile)@left / max(readWave(soundfile))
  # autoCorrelation: pre-calculated autocorrelation of a frame supplied by function analyzeSound ()
  
  # just some initializing
  HNR = NA 
  pitch_array = data.frame ('pitchCand'=NA,'pitchAmpl'=NA,'source'=NA, stringsAsFactors=F, row.names=NULL)
  pitchAutocor_array = pitchCep_array = pitchSpec_array = dom_array = data.frame ('pitchCand'=numeric(),'pitchAmpl'=numeric(),'source'=character(), stringsAsFactors=F, row.names=NULL)
  summaries = data.frame ('HNR'=NA, 'dom'=NA, 'peakFreq'=NA, 'peakFreq_cut'=NA, 'meanFreq'=NA, 'quartile25'=NA, 'quartile50'=NA, 'quartile75'=NA, 'specSlope'=NA)
  
  ### DESCRIPTIVES
  meanSpec = data.frame ('freq'=1000*as.numeric(names(frame)), 'amp'=frame)
  amplitude = sum(frame)
  peakFreq = meanSpec$freq [which.max(frame)] # absolute peak - peak frequency
  meanFreq = meanSpec$freq[min(which(cumsum(frame) > amplitude/2))] # spectral center of mass
  
  # Cut spectral band from pitch_floor to cutoff_dom Hz
  meanSpec_cut = meanSpec [meanSpec$freq >pitch_floor & meanSpec$freq < cutoff_dom, ] # Above 5-6 kHz or so, spectral energy depends too much on the original sampling rate, noises etc. Besides, those frequencies are not super relevant to human vocalizations in any case. So we cut away all info above 5 kHz before we calculate quartiles of spectral energy
  peakFreq_cut = meanSpec_cut$freq[which.max(frame)] # peak frequency under 5 kHz
  amplitude_cut = sum(meanSpec_cut$amp)
  quartile25 = meanSpec_cut$freq [min(which( cumsum(meanSpec_cut$amp) >= 0.25 * amplitude_cut ))] # first quartile of spectral energy distribution in the band from pitch_floor to cutoff_dom kHz
  quartile50 = meanSpec_cut$freq [min(which( cumsum(meanSpec_cut$amp) >= 0.5 * amplitude_cut ))] # second quartile (same as mean freq within this spectral band)
  quartile75 = meanSpec_cut$freq [min(which( cumsum(meanSpec_cut$amp) >= 0.75 * amplitude_cut ))] # third quertile. Note: half the energy in the band from pitch_floor to cutoff_dom kHz lies between quartile25 and quartile75
  specSlope = summary(lm(amp~freq, data=meanSpec_cut))$coef[2,1]
  
  # get lowest dominant frequency bands
  dom = NA
  frame = frame/max(frame) # plot (frame, type='l')
  frame_zoo = as.zoo(frame)
  bin = samplingRate/2/length(frame) # the width of one bin in spectrogram, in Hz (~20 Hz for 44100 Hz with 50 ms window and zp=0)
  width = 2*ceiling(12 * 20/bin / 2)-1 # width of smoothing interval, chosen to be always ~220 Hz, regardless of bin, but an uneven number. Chosen by iterative optimization to produce dom values as close as possible to manually verified pitch (see file domTraining.R)
  temp = rollapply(frame_zoo, width = width, align='center', function(x) isCentral.localMax(x, threshold=dom_threshold))
  idx = index(temp)[coredata(temp)]
  pitch_floor_idx = which (as.numeric(names(frame))>pitch_floor/1000)[1]
  
  if (length(idx)>0){
    dom = as.numeric(names(frame)[idx[idx>pitch_floor_idx][1]])*1000 # lowest dominant freq band - we take the first frequency in the spectrum at least /dom_threshold/ % of the amplitude of peak frequency, but high enough to be above pitch_floor
    dom_array = data.frame ('pitchCand' = dom, 'pitchAmpl' = frame[idx[1]], 'source' = 'dom', stringsAsFactors=F, row.names=NULL)
  } 
  
  ### PITCH TRACKING  
  
  ## PITCH_AUTOCORRELATION. Try to detect pitch with autocorrelation function (time domain analysis). Modified PRAAT's algorithm (P. Boersma)
  # autoCorrelation = autocorBank[,21]
  a = data.frame ('freq' = as.numeric(names(autoCorrelation)), 'amp' = autoCorrelation); rownames(a)=NULL 
  a = a[a$freq>pitch_floor & a$freq<pitch_ceiling,,drop=F] # plot(a[,2], type='l')
  HNR = max(a$amp) # HNR is here defined as the maximum autocorrelation within the specified pitch range. It is also measured for the frames which are later classified as unvoiced (i.e. HNR can be <voiced_threshold)
  
  # find peaks in the corrected autocorrelation function
  a_zoo = as.zoo(a$amp) 
  temp = rollapply(a_zoo, width=7, align='center', function(x) isCentral.localMax(x, threshold=voiced_threshold_autocor)) # width=7 chosen by optimization, but it doesn't make that much difference anyhow
  idx = index(temp)[coredata(temp)]
  autocorPeaks = a[idx,]
  
  if (nrow(autocorPeaks)>0) { # if some peaks are found...
    # we are only interested in frequencies above half of the best candidate (b/c otherwise we get false subharmonics)
    bestFreq = autocorPeaks$freq[which(autocorPeaks$amp>0.975*max(autocorPeaks$amp))[1]]
    # bestFreq = autocorPeaks$freq[which.max(autocorPeaks$amp)] 
    if (!is.na(bestFreq)) { 
      autocorPeaks = try(autocorPeaks[autocorPeaks$freq > bestFreq/1.8,,drop=F], silent=T) # otherwise we get false subharmonics
      autocorPeaks = try(autocorPeaks[order(autocorPeaks$amp, decreasing = T),,drop=F], silent=T)
    }         
    if (class(autocorPeaks)!='try-error') {
      if (nrow(autocorPeaks)>0) { # if some peaks satisfy all criteria, return them:
        pitchAutocor_array = data.frame ('pitchCand' = autocorPeaks [1:min(nrow(autocorPeaks),maxNoCands),1], # save n candidates of pitchAutocor, convert to Hz
                                         'pitchAmpl' = autocorPeaks[1:min(nrow(autocorPeaks),maxNoCands),2], # save amplitudes corresponding to each pitchAutocor candidate
                                         'source' = 'autocor', stringsAsFactors=F, row.names=NULL)
      }
    } 
  }
  # very occasionally HNR can be calculated as 1.01 etc. To prevent this nonsense:
  if (!is.na(HNR) & HNR >= 1) {HNR=0.9999} # analogous to 40 dB
  ## END OF PITCH_AUTOCORRELATION
  
  
  ## PITCH_CEPSTRAL. Try to detect pitch by using ~cepstrum (fft of the spectrum). See http://www.phon.ucl.ac.uk/courses/spsci/matlab/lect10.html; modified (see file cepstrum-2.R)
  if (zpCep < length(frame)){
    frameZP = frame
  } else {
    frameZP = c (frame, rep(0,(zpCep-length(frame))))
  }
  
  cepstrum = abs(fft(frameZP)) # plot(frameZP, type='l') # fft of fft, whatever you call it - cepstrum or smth else
  cepstrum = cepstrum / max(cepstrum) # plot (cepstrum, type='l')
  b = cbind (1:(length(cepstrum)%/%2), cepstrum[1:(length(cepstrum)%/%2)]) # plot (b, type='l')
  b[,1] = samplingRate / (1:nrow(b)) /2 * (length(frameZP)/length(frame)) # NB: divide by 2 because it's another fft, not reverse fft (cf. pitchAutocor)
  highestRelevantPeak = which(b[,1] < pitch_ceiling)[1] # if we are interested in high frequencies, we need to be able to detect closely spaced peaks in the cepstrum, thus we choose a relatively narrow width of the window for finding peaks. And vice versa. If we took a width double the highestRelevantPeak, we would fail to find the very first peak (cepstrum starts with 1, like autocor), so we just take width=highestRelevantPeak
  b = b[b[,1]>pitch_floor & b[,1]<pitch_ceiling, ]
  
  # find peaks
  a_zoo = as.zoo(b[,2]) 
  temp = rollapply(a_zoo, width=highestRelevantPeak, align='center', function(x) isCentral.localMax(x, threshold=voiced_threshold_cep))
  idx = index(temp)[coredata(temp)] 
  
  absCepPeak = try (idx[which.max(b[idx,2])], silent = T)
  if (class(absCepPeak)!='try-error') {idx = idx[b[idx,1] > b[absCepPeak,1]/1.8 ]} # to avoid false subharmonics
  # plot (b, type='l'); points(b[idx,1], b[idx,2])
  acceptedCepPeak = idx[which.max(b[idx,2])] # the highest cepstrum peak within pitch range
  cepstrumPeaks = data.frame ('freq' = b[acceptedCepPeak,1], 'amp' = b[acceptedCepPeak,2])
  
  if (nrow(cepstrumPeaks)>0) { # if some peaks are found...
    pitchCep_array = data.frame ('pitchCand' = cepstrumPeaks$freq[1],
                                 'pitchAmpl' = cepstrumPeaks$amp[1],
                                 'source' = 'cepstrum', stringsAsFactors=F, row.names=NULL)
    # because cepstrum really stinks for frequencies above ~1 kHz, mostly picking up formants or just plain noise, we discount confidence in high-pitch cepstral estimates
    pitchCep_array$pitchAmpl = pitchCep_array$pitchAmpl / (1+5/(1+exp(14-3*log2(pitchCep_array$pitchCand/pitch_floor)))) # visualization: a = seq(pitch_floor, pitch_ceiling, length.out=100); b = 1+5/(1+exp(14-3*log2(a/pitch_floor))); plot (a,b, type='l')
  } 
  ## END OF PITCH_CEPSTRAL
  
  
  ## PITCH_SPECTRAL. Try to detect pitch by finding peaks in the spectrum (frequency domain analysis, ~ modified BaNa algorithm)
  width = 2*ceiling((widthSpec/bin+1) * 20/bin / 2)-1 # to be always ~100 Hz, regardless of bin, but an odd number
  if (is.na(HNR)){
    specPitchThreshold = specPitchThreshold_nullNA # if HNR is NA, the sound is probably a mess, so we play safe by only looking at very strong harmonics
  } else {
    specPitchThreshold = specPitchThreshold_nullNA - specPitchThreshold_nullNA*HNR*slopeSpec  # for noisy sounds the threshold is high to avoid false sumharmonics etc, for tonal sounds it is low to catch weak harmonics
  }
  temp = rollapply(frame_zoo, width = width, align='center', function(x) isCentral.localMax(x, threshold=specPitchThreshold)) # plot(frame_zoo, type='l')
  idx = index(temp)[coredata(temp)] 
  specPeaks = data.frame ('freq' = as.numeric(names(frame)[idx])*1000, 'amp' = frame[idx])
  
  if (nrow(specPeaks)==1) {
    if (specPeaks[1,1]<pitch_ceiling & specPeaks[1,1]>pitch_floor){
      pitchSpec = specPeaks[1,1]
      pitchAmpl = pitchSpec_only_peak_weight
      pitchSpec_array = data.frame('pitchCand'=pitchSpec, 'pitchAmpl'=pitchSpec_only_peak_weight, 'source'='spec', stringsAsFactors=F, row.names=NULL)
    }
  } else if (nrow(specPeaks)>1) { # 
    specPeaks = specPeaks [1:min(5,nrow(specPeaks)),] # analyze five lowest harmonics
    
    # A modified version of BaNa algorithm follows
    seq1 = 2:nrow(specPeaks)
    seq2 = 1:(nrow(specPeaks)-1)
    n = length(seq1)*length(seq2)
    temp = data.frame ('harmonicA' = rep(0,n), 'harmonicB' = rep(0,n), 'AtoB_ratio' = rep(0,n))
    counter = 1
    for (i in seq1) {
      for (j in seq2) {
        temp$harmonicA[counter] = i
        temp$harmonicB[counter] = j
        temp$AtoB_ratio[counter] = specPeaks$freq[i] / specPeaks$freq[j] # get ratios of discovered harmonics to each other
        counter = counter+1
      }
    }
    temp = temp[temp$harmonicA>temp$harmonicB,]
    
    pitchCand = numeric()
    for (i in 1:nrow(temp)) {
      pitchCand = c(pitchCand, as.numeric(specPeaks$freq[temp$harmonicB[i]] / ratios$divide_lower_by [temp$AtoB_ratio[i]>ratios$value_low & temp$AtoB_ratio[i]<ratios$value_high])) # for each ratio that falls within the limits specified outside this function in a dataframe called "ratios", calculate the corresponding pitch. If several ratios suggest the same pitch, that's our best guess
    }
    
    # add pitchCand based on the most common distances between harmonics
    # pitchCand = c(pitchCand, diff(specPeaks[,1]))
    
    pitchCand = sort (pitchCand [pitchCand>pitch_floor & pitchCand<pitch_ceiling])
    if (length(pitchCand)>0) {
      pitchSpec_array = data.frame('pitchCand'=pitchCand, 'specAmplIdx'=1, 'source'='spec', stringsAsFactors=F, row.names=NULL)
      c = 1
      while (c+1 <= nrow(pitchSpec_array)){
        if (abs(log2(pitchSpec_array$pitchCand[c]/pitchSpec_array$pitchCand[c+1])) < 1/12){ # merge cands within 1 semitone into one "super-candidate" and give this new super-candidate a certainty boost
          pitchSpec_array$specAmplIdx[c] = pitchSpec_array$specAmplIdx[c] + 1
          pitchSpec_array$pitchCand[c] = mean ( c(pitchSpec_array$pitchCand[c], pitchSpec_array$pitchCand[c+1]))
          pitchSpec_array = pitchSpec_array[-(c+1),]
        } else {
          c = c+1
        }
      }
      pitchSpec_array$pitchAmpl = pitchSpec_only_peak_weight + (1/(1+exp(-(pitchSpec_array$specAmplIdx-1)))-0.5)*2 * (1-pitchSpec_only_peak_weight) # normalization. Visualization: a=1:15; b = pitchSpec_only_peak_weight + (1/(1+exp(-(a-1)))-0.5)*2 * (1-pitchSpec_only_peak_weight); plot(a,b,type='l')
      pitchSpec_array = pitchSpec_array[order(pitchSpec_array$pitchAmpl, decreasing=T), c('pitchCand','pitchAmpl','source')]
    } 
  }
  if (sum(!is.na(pitchSpec_array))>0) {
    pitchSpec_array = pitchSpec_array[pitchSpec_array$pitchAmpl>voiced_threshold_spec,, drop=F]
    pitchSpec_array = pitchSpec_array[1:min(nrow(pitchSpec_array),maxNoCands),]  # how many pitchSpec candidates to use (max)
  }
  ## END OF PITCH_SPECTRAL
  
  ## merge all pitch candidates in a single list
  pitch_array = rbind (pitchAutocor_array, pitchCep_array, pitchSpec_array, dom_array)
  if (nrow(pitch_array)>0){
    pitch_array[, 1:2] = apply (pitch_array[, 1:2], 2, function(x)as.numeric (x) ) # otherwise become characters after rbind 
  }
  if (nrow(pitch_array[pitch_array$source=='dom',])>0 & !is.na(HNR)){
    pitch_array$pitchAmpl[pitch_array$source=='dom'] = 1/(1+exp(3*HNR-1)) # dom is worth more for noisy sounds, but its weight approaches ~0.2 as HNR approaches 1 (NB: this is before HNR is converted to dB). Visualization: a = seq(0,1,length.out=100); b = 1/(1+exp(3*a-1)); plot (a,b, ylim=c(0,1))
  }
  
  return (list ('pitch_array'=pitch_array, 'summaries'=data.frame(HNR=HNR, dom=dom, peakFreq=peakFreq, peakFreq_cut=peakFreq_cut, meanFreq=meanFreq, quartile25=quartile25, quartile50=quartile50, quartile75=quartile75, specSlope=specSlope)))
}

# for debugging:
# n = which(as.numeric(colnames(s))>100) [1]  # to find the frame number for time in ms
# frame = s[,n]; autoCorrelation = autocorBank[,n]
# a = analyzeFrame (frame=frame, autoCorrelation=autoCorrelation, maxNoCands=2, pitch_ceiling=15000)




#
# **************************************
# F R O N T - E N D    F U N C T I O N S
# **************************************
# 

spectro_denoised = function (soundfile, frameBank=NULL, duration=NULL, samplingRate=NULL, contrast = 0, brightness=2, windowLength = 50, step = 15, ylim = c(0,5), wn='gaussian', zp = 0, main=NULL, plot=T, returnZ=F) {
  ## Produces a black & white spectrogram. Input can be full path to a .wav file, a vector of amplitudes with specified samplingRate, or a previously saved bank of individual frames as produced by function getFrameBank() called from analyzeSound()
  
  # contrast (0 to +inf): how much noise to remove. 0 = original sound, 1 = remove as much noise as there is in the noisiest frame, 2 = remove x2 as much noise as in the noisiest frame, etc (very large contrast produces an almost-empty spectrogram)
  # brightness (0 to +inf): how much to "lighten" the image. 1 = original sound, >1 makes the spectrogram lighter and <1 darker by re-scaling intensities (thus without affecting the darkest regions)
  # windowLength in ms
  # wn - window type: gaussian, hanning, hamming, bartlett, rectangular, blackman, flattop
  # step - step between analysis frames, in ms. Shorter steps yield better temporal resolution but require longer processing time
  # zp - zero padding specified as the desired final window length in points, including zeroes(! ≠seewave). If zp < window length recalculated in points, nothing happens. If zp > window length, spectral resolution may improve without sacrificing time resolution (NB: slows down the analysis!)
  
  # import audio
  if (class(soundfile)=='character'){
    sound_wav = readWave (soundfile)
    samplingRate = sound_wav@samp.rate
    windowLength_points = floor (windowLength/1000 * samplingRate / 2) * 2
    # windowLength_points = 2^round (log(windowLength * samplingRate /1000)/log(2), 0) # to ensure that the window length in points is a power of 2, say 2048 or 1024
    sound = sound_wav@left 
    duration = length(sound) / samplingRate
    frameBank = getFrameBank (sound=sound, samplingRate=samplingRate, windowLength_points=windowLength_points, step=step, zp=zp, wn=wn, filter=NULL) 
  } else if (class(soundfile)=='numeric' & length(soundfile)>1){
    if (is.null(samplingRate)){
      stop ('Please specify samplingRate, eg 44100')
    } else {
      duration = length(sound) / samplingRate
      windowLength_points = floor(windowLength/1000 * samplingRate / 2) * 2
      frameBank = getFrameBank (sound=soundfile, samplingRate=samplingRate, windowLength_points=windowLength_points, step=step, zp=zp, filter=NULL)
    }
  } 
  if (class(frameBank)!='matrix') {
    stop('Input format not recognized. Please provide path to .wav file, a vector of amplitudes plus samplingRate, or a preprocessed frameBank')
  }
  
  windowLength_points = floor(windowLength/1000 * samplingRate / 2) * 2
  z <- apply(frameBank, 2, function(x) abs(fft(x)[1:(floor(nrow(frameBank)/2))]) )  # fft of each frame
  X <- seq(0, duration*1000, length.out = ncol(z)) # time stamp
  Y <- seq(0, (samplingRate/2) - (samplingRate/windowLength_points), length.out = nrow(z))/1000 # frequency stamp
  rownames(z) = Y; colnames(z)=X
  Z = t(z)
  
  # removing noise
  if (contrast>0){ # can get slow, especially with zero padding
    global_max = max ( apply(Z,1,sum) )
    noisiest_frame = Z [which.min (apply(Z, 1, function(x)sum(x)/global_max/getEntropy(x))),] # frame closest to "pure noise" - queitest and with largest entropy
    m1 = max(noisiest_frame) * contrast # the amount of noise to remove
    Z1 = ifelse (Z <= m1, 1e-10, Z-m1) # make all values under threshold vanishingly small, subtract noise from those above threshold
  } else {
    Z1 = Z
  }
  Z1 = Z1 / max(Z1)
  Z1[Z1==0] = 1e-10
  
  # spectrogram of the modified fft
  if (plot){
    filled.contour.modif2 (x = X, y = Y, z = log(Z1)*brightness, levels = seq(min(log(Z1)), max(log(Z1)), length=30), color.palette = function (x) gray(seq(from = 1, to = 0, length = x)), ylim = ylim, main=main) # , xlab='Time, ms', ylab = 'Frequency, kHz'
  }
  if (returnZ){return (t(Z1))} # or return(z) for unchanged spectrogram (w/o denoising)
}

## Example of use:
# spectro_denoised ('myfile.wav', contrast=0, brightness=1, wn='gaussian', ylim=c(0,10), step=50)



analyzeSound = function (soundfile, silence=0.03, entropyThreshold=0.9, windowLength=50, wn='gaussian', step=25, zp=0, zpCep=2^13, pitch_floor=75, pitch_ceiling=3500, maxPitchCands=4, voiced_threshold_autocor=0.75, voiced_threshold_cep=0.45, voiced_threshold_spec=0.5, specPitchThreshold_nullNA=0.5, slopeSpec=0.1, widthSpec=150, pitchSpec_only_peak_weight=0.51, priorMean=NA, priorSD=NA, cutoff_dom=6000, dom_threshold=0.1, shortestSyllable=20, shortestBreak=60, interpolWindow=3, interpolTolerance=0.3, interpolCert=0.3, certWeight=.5, runSnake=T, snakeSmoothingStep=0.05, snakeIterMultiplier=2, plotSnake=F, smooth=T, smoothIdx=1, plot=T, savePlot=F, savePlotPath='~/Downloads/temp/', contrast = 0, brightness=2, ylim=c(0,5), smooth_vars=c('pitch','dom')) {
  ## Acoustic analysis of a single sound file
  
  # soundfile - full path to a .wav file
  # windowLength - in ms 
  # wn - window type: gaussian, hanning, hamming, bartlett, rectangular, blackman, flattop
  # voicedthreshold: anything below has no pitch, but I do calculate HNR for frames classified as unvoiced (a matter of taste). Specified separately for each pitch detection algorithm (pitchAutocor, pitchCep, pitchSpec)
  # pitch_floor, pitch_ceiling - in Hz
  # maxPitchCands: how many pitch candidates to return (at most). NB: only one dom and one pitchCep is returned, so the remaining candidates come from the autocorrelation and spectral pitch candidates
  # zp: zero-pad the spectrogram? zpCep: zero-pad the spectrum used for cepstral pitch detection? zpCep improves the precision of cepstral pitch detection quite noticeably
  # silence - frames with mean abs amplitude below silence threshold are not analyzed. NB: adaptive! The actual silence threshold may be higher depending on the quietest frame, but it will not be lower than this specified number
  # dom_threshold: to find the lowest dominant frequency band, we do short term FT and take the lowest frequency with amplitude at least dom_threshold
  # cutoff_dom: do not consider frequencies above cutoff_dom when calculating the lowest dominant frequency band (recommended if the original sampling rate varies, as in the training corpus)
  # shortestSyllable: the smallest length of a voiced segment (ms) that constitutes a syllable (shorter segments will be replaced by NA as if unvoiced)
  # shortestBreak: the smallest gap between voiced syllables (ms) that means they shouldn't be merged into one voiced syllable
  # smoothing - if TRUE, contours of the specified variables (smooth_vars) are smoothed. Control the amount of smoothing with smoothIdx
  # interpol pars specify the behavior of interpolation algorithm (post-processing of pitch candidates); see code for details above this and other parameters regulating postprocessing
  # certWeight: specifies the weight of certainty in pitch estimates vs F0 jumps (see function 'cost'). High certWeight: we mostly pay attention to our certainty in particular pitch candidates; low certWeight: we are more concerned with avoiding rapid pitch fluctuations in our contour
  # pitchSpec_only_peak_weight: if only one pitchSpec candidate is found, its weight (certainty) is taken to be ... (0 to 1 - this number is used for comparing two pitch estimates - on time/frequency domain). Also used for deciding which pitch to use - autocor or spec
  # prior: specify the mean and sd of gamma distribution describing our prior knowledge about the most likely pitch values for this file (defaults to NA)
  
  sound_wav = readWave (soundfile)
  samplingRate = sound_wav@samp.rate
  windowLength_points = floor(windowLength/1000 * samplingRate / 2) * 2
  # windowLength_points = 2^round (log(windowLength * samplingRate /1000)/log(2), 0) # to ensure that the window length in points is a power of 2, say 2048 or 1024
  sound = sound_wav@left 
  if (min(sound)>0) {sound = scale(sound)}
  sound = sound / max( abs(max(sound)), abs(min(sound))) # normalize to range from no less than -1 to no more than +1
  duration = length(sound) / samplingRate
  maxNoCands = (maxPitchCands-2) %/% 2 # one for dom, one for pitchCep, the rest shared between pitchAutocor and pitchSpec
    
  # Set up filter for calculating pitchAutocor
  filter = ftwindow_modif (2*windowLength_points, wn=wn) # plot(filter, type='l')
  powerSpectrum_filter = abs(fft(filter))^2 
  autoCorrelation_filter = abs(fft(powerSpectrum_filter, inverse=T))^2
  autoCorrelation_filter = autoCorrelation_filter[1:windowLength_points]
  autoCorrelation_filter = autoCorrelation_filter / max(autoCorrelation_filter) # plot(autoCorrelation_filter, type='l')
  
  # fft and acf per frame
  if (savePlot) {
    plot = T
    plotname = tail (unlist (strsplit(soundfile, '/')),n=1)
    plotname = substring (plotname, first=1, last=(nchar(plotname)-4))
    jpeg(file = paste0 (savePlotPath, plotname, ".jpg"), 1200,800)
  }
  frameBank = getFrameBank (sound=sound, samplingRate=samplingRate, windowLength_points=windowLength_points, wn=wn, step=step, zp=zp, filter=NULL)
  s = spectro_denoised (frameBank=frameBank, soundfile=NULL, duration=duration, samplingRate=samplingRate, ylim=ylim, windowLength=windowLength, zp=zp, wn=wn, contrast=contrast, brightness=brightness, step=step, main=tail(unlist(strsplit(soundfile,'/')),1), plot=plot, returnZ=T)
  autocorBank = apply (frameBank, 2, function(x) acf(x, windowLength_points, plot=F)$acf/autoCorrelation_filter)  # plot (windowBank[,5],type='l')
  rownames(autocorBank) = samplingRate / (1:nrow(autocorBank)) 
  
  # calculate amplitude of each frame
  myseq = seq(1, (length(sound)-windowLength_points), length.out=ncol(s))
  ampl = apply (as.matrix(1:ncol(s)), 1, function(x){
    sqrt( mean(sound[myseq[x] : (myseq[x]+windowLength_points)]^2) ) # perceived intensity - root mean square of amplitude
  })
  silence_threshold = max (silence, min(ampl))
  
  # calculate entropy of each frame
  rowLow = which(as.numeric(rownames(s))>0.05)[1] # 50 Hz
  rowHigh = which(as.numeric(rownames(s))>6)[1] # 6000 Hz
  entropy = apply (as.matrix(1:ncol(s)), 1, function(x){
    getEntropy(s[rowLow:rowHigh,x]) # entropy is measured within the most relevant vocal range only: 50 to 6000 Hz
  })
  cond = ampl>silence & entropy<entropyThreshold # if the frame is too quiet or too noisy, we will not analyze it
  
  # spectral analysis of each frame from fft
  frameInfo = rep(list(list('pitch_array' = data.frame ('pitchCand'=NA,'pitchAmpl'=NA,'source'=NA, stringsAsFactors=F, row.names=NULL), 'summaries' = data.frame ('HNR'=NA, 'dom'=NA, 'peakFreq'=NA, 'peakFreq_cut'=NA, 'meanFreq'=NA, 'quartile25'=NA, 'quartile50'=NA, 'quartile75'=NA, 'specSlope'=NA))), ncol(s)) # set up an empty nested list to save values in - this enables us to analyze only the non-silent and not-too-noisy frames but still have a consistently formatted output
  for (i in (1:ncol(s))[cond]){ # for each frame that satisfies our condition, do spectral analysis (NB: we do NOT analyze frames that are too quiet or have very high entropy, so we only get NA's for those frames, no meanFreq, dom etc!)
    frameInfo [[i]] = analyzeFrame (frame=s[,i], autoCorrelation=autocorBank[,i], samplingRate=samplingRate, windowLength_points=windowLength_points, wn=wn, zp=zp, zpCep=zpCep, cutoff_dom=cutoff_dom, voiced_threshold_autocor=voiced_threshold_autocor, voiced_threshold_cep=voiced_threshold_cep, voiced_threshold_spec=voiced_threshold_spec, specPitchThreshold_nullNA=specPitchThreshold_nullNA, slopeSpec=slopeSpec, widthSpec=widthSpec, pitch_floor=pitch_floor, pitch_ceiling=pitch_ceiling, silence=silence, dom_threshold=dom_threshold, pitchSpec_only_peak_weight=pitchSpec_only_peak_weight, autoCorrelation_filter=autoCorrelation_filter, maxNoCands=maxNoCands)
  }
  
  # Store the descriptives provided by function analyzeFrame in a dataframe
  summaries = lapply(frameInfo, function(y) y[['summaries']])
  result = do.call("rbind", summaries)
  result = as.data.frame (result)
  result$ampl = ampl
  result$entropy = entropy
  result$time = round ( seq(windowLength_points/2/samplingRate, duration, length.out=nrow(result))*1000,0)
  result$duration=duration
  rownames(result) = 1:nrow(result) # otherwise weird row names - bad for debugging
  
  # extract and prepare pitch candidates for the pathfinder algorithm
  pitch_list = lapply(frameInfo, function(y) y[['pitch_array']])
  pitchCands = lapply(pitch_list, function(y) as.data.frame(t(y[['pitchCand']])))
  pitchCands = t(rbind.fill(pitchCands)) # a matrix of pitch candidates per frame
  pitchCert = lapply(pitch_list, function(y) as.data.frame(t(y[['pitchAmpl']])))
  pitchCert=t(rbind.fill(pitchCert)) # a matrix of our certainty in pitch candidates
  pitchSource = lapply( pitch_list, function(y) as.data.frame(t(y[['source']]), stringsAsFactors=F) ) # NB: without StringsAsFactors=F, the first row becomes "1" because of wrong NA recognition
  pitchSource=t(rbind.fill(pitchSource)) # a matrix of the sources of pitch candidates
   
  # PRIOR for adjusting the estimated pitch certainties. For ex., if primarily working with speech, we could prioritize pitch candidates in the expected pitch range (100-1000 Hz) and dampen candidates with very high or very low frequency as unlikely but still remotely possible in everyday vocalizing contexts (think a soft pitch ceiling)
  if (!is.na(priorMean) & !is.na(priorSD)){
    prior_normalizer = max(dgamma (log2(seq(pitch_floor,pitch_ceiling,length.out=100)), shape=log2(priorMean)^2/log2(priorSD)^2, rate=log2(priorMean)/log2(priorSD)^2))
    pitchCert = pitchCert * dgamma (log2(pitchCands), shape=log2(priorMean)^2/log2(priorSD)^2, rate=log2(priorMean)/log2(priorSD)^2) / prior_normalizer
  }
  
  # divide the file into continuous voiced syllables
  analFrames = findVoicedSegments (pitchCands, shortestSyllable=shortestSyllable, shortestBreak=shortestBreak, step=step, samplingRate=samplingRate)
  
  # for each syllable, impute NA's and find a nice path through pitch candidates 
  pitchFinal = rep (NA, ncol(pitchCands))
  if (nrow(analFrames)>0){ # if we have found at least one putatively voiced syllable
    for (syl in 1:nrow(analFrames)) { 
      myseq = analFrames$segmentStart[syl] : analFrames$segmentEnd[syl]     
      # compute the optimal path through pitch candidates
      pitchFinal[myseq] = pathfinder (pitchCands=pitchCands[,myseq], pitchCert=pitchCert[,myseq], certWeight=certWeight, interpolWindow=interpolWindow, interpolTolerance=interpolTolerance, interpolCert=interpolCert, runSnake=runSnake, snakeSmoothingStep=snakeSmoothingStep, snakeIterMultiplier=snakeIterMultiplier, plotSnake=plotSnake)
    }
  }
   
  # save optimal pitch track and the best candidates separately for autocor, cepstrum and spectral
  result$pitch <- pitchFinal # optimal pitch track
  result$pitchAutocor <- as.numeric ( lapply (pitch_list, function(x)x$pitchCand[x$source=='autocor'] [which.max(x$pitchAmpl[x$source=='autocor'])]) )
  result$pitchCepstrum <- as.numeric ( lapply (pitch_list, function(x)x$pitchCand[x$source=='cepstrum'] [which.max(x$pitchAmpl[x$source=='cepstrum'])]) )
  result$pitchSpec <- as.numeric ( lapply (pitch_list, function(x)x$pitchCand[x$source=='spec'] [which.max(x$pitchAmpl[x$source=='spec'])]) )
  
  ## Median smoothing of specified contours (by default pitch & dom)
  if (smooth) { 
    points_per_sec = nrow(result) / duration
    smoothing_ww = round (smoothIdx * points_per_sec/10, 0) # smoothIdx of 1 means that smoothing window is ~100 ms
    smoothing_threshold = 4 / smoothIdx # the larger smoothIdx, the heavier the smoothing (lower tolerance threshold before values are replaced by median over smoothing window). smoothIdx of 1 gives smoothing_threshold of 4 semitones
    result[smooth_vars] = medianSmoother (result[smooth_vars], smoothing_ww=smoothing_ww, smoothing_threshold=smoothing_threshold)
  }
  
  ## Having decided upon the pitch for each frame, we save certain measurements only for voiced frames (with non-NA pitch)
  result$quartile25 = ifelse (!is.na(result$pitch), result$quartile25, NA)
  result$quartile50 = ifelse (!is.na(result$pitch), result$quartile50, NA)
  result$quartile75 = ifelse (!is.na(result$pitch), result$quartile75, NA)
  result$voiced = ifelse(!is.na(result$pitch), T, F)
  result$ampl_voiced = ifelse (!is.na(result$pitch), result$ampl, NA)
  
  ## Calculate the % of energy in harmonics based on the final pitch estimates
  threshold = 1.25 * result$pitch / 1000
  result$harmonics = apply (matrix(1:ncol(s)), 1, function(x)ifelse(is.na(threshold[x]), NA, sum (s[as.numeric(rownames(s))>threshold[x], x]) / sum (s[,x])))
  
  ## Convert HNR and harmonics to dB
  result$HNR = to_dB(result$HNR)
  result$harmonics = to_dB(result$harmonics)
  
  ## Plotting and harmonics based on the final pitch contour
  if (plot) {
    mylevels = c('autocor','cepstrum','spec','dom')
    mycols = c('green','violet','red','orange')
    mypch = c(16,7,2,3)
    # pitchSource_1234 = apply(pitchSource, 2, function(x)match(x,mylevels))
    pitchSource_1234 = matrix(match(pitchSource,mylevels),ncol=ncol(pitchSource))
    for (r in 1:nrow(pitchCands)){
      points (result$time, pitchCands[r,]/1000, col=mycols[pitchSource_1234[r,]], pch=mypch[pitchSource_1234[r,]], cex=pitchCert[r,]*2)
    }
    lines (result$time, result$pitch/1000, col='blue', lwd=3)
  }
  if (savePlot) {dev.off()} 
 
  result = result[c('duration','time','voiced','ampl','ampl_voiced','entropy','HNR','dom','meanFreq','peakFreq','peakFreq_cut','pitch','pitchAutocor','pitchCepstrum','pitchSpec','quartile25','quartile50','quartile75','specSlope','harmonics')]
  return (result)
}

# example of use: a = analyzeSound ('soundfile.wav', plot=T)



analyzeFolder = function (folder, silence=0.03, entropyThreshold=0.9, windowLength=50, wn='bartlett', step=25, zp=0, zpCep=2^13, pitch_floor=75, pitch_ceiling=3500, maxPitchCands=4, voiced_threshold_autocor=0.75, voiced_threshold_cep=0.45, voiced_threshold_spec=0.5, pitchSpec_only_peak_weight=0.51, specPitchThreshold_nullNA=0.5, slopeSpec=0.1, widthSpec=150, priorMean=NA, priorSD=NA, cutoff_dom=6000, dom_threshold=0.1, shortestSyllable=30, shortestBreak=40, interpolWindow=2, interpolTolerance=0, interpolCert=0.1, certWeight=.5, runSnake=T, snakeSmoothingStep=0.07, snakeIterMultiplier=2, plotSnake=F, smooth=T, smoothIdx=2, plot=T, savePlot=F, savePlotPath='~/Downloads/temp/', contrast = 0, brightness=2, ylim=c(0,5), smooth_vars=c('pitch','dom'), verbose=T) {
  # analyzes all .wav files in a folder. Returns a dataframe with acoustic descriptors and (optionally) saves the spectrograms with pitch contours
  
  ptm <- proc.time()   
  filenames <- list.files(folder, pattern="*.wav", full.names=TRUE)       
  filesizes = apply (as.matrix(filenames), 1, function(x)file.info(x)$size) # in order to provide more accurate estimates of time to completion
  vars = c('ampl','ampl_voiced','entropy','HNR','dom','meanFreq','peakFreq','peakFreq_cut','pitch','pitchAutocor','pitchCepstrum','pitchSpec','quartile25','quartile50','quartile75','specSlope','harmonics') 
  out = as.data.frame (matrix(ncol=3+2*length(vars), nrow=length(filenames)))  
  colnames(out)[c(1:3)] = c('file','duration','voiced')
  for (c in 1:length(vars)){
    colnames(out)[3+2*c-1] = paste0(vars[c],'_','median') # specify how to summarize pitch etc values for each frame within each file - save median, sd, ... "3+2*c-1": "3" because of file/dur/voiced above, "+2*c" because for each acoustic variable, we save median and sd
    colnames(out)[3+2*c] = paste0(vars[c],'_','sd')
  }
  myseq = (1:length(vars))+3 # which columns in the output of pitch_per_sound to save as median + sd
  
  for (i in 1:length(filenames)) {
    temp = analyzeSound (filenames[i], silence=silence, entropyThreshold=entropyThreshold, windowLength=windowLength, wn=wn, step=step, voiced_threshold_autocor=voiced_threshold_autocor, voiced_threshold_cep=voiced_threshold_cep, voiced_threshold_spec=voiced_threshold_spec, specPitchThreshold_nullNA=specPitchThreshold_nullNA, slopeSpec=slopeSpec, widthSpec=widthSpec, zp=zp, zpCep=zpCep, pitch_floor=pitch_floor, pitch_ceiling=pitch_ceiling, maxPitchCands=maxPitchCands, pitchSpec_only_peak_weight=pitchSpec_only_peak_weight, priorMean=priorMean, priorSD=priorSD, cutoff_dom=cutoff_dom, dom_threshold=dom_threshold, shortestSyllable=shortestSyllable, shortestBreak=shortestBreak, interpolWindow=interpolWindow, interpolTolerance=interpolTolerance, interpolCert=interpolCert, certWeight=certWeight, runSnake=runSnake, snakeSmoothingStep=snakeSmoothingStep, snakeIterMultiplier=snakeIterMultiplier, plotSnake=plotSnake, smooth=smooth, smoothIdx=smoothIdx, plot=plot, savePlot=savePlot, savePlotPath=savePlotPath, contrast=contrast, brightness=brightness, ylim=ylim, smooth_vars=smooth_vars)
    out[i,1] = tail (unlist (strsplit(filenames[i], '/')),n=1)
    out[i,2] = temp[1,'duration'] # duration, ms
    out[i,3] = mean (temp[,'voiced']) # proportion of voiced frames
    
    for (v in 1:length(myseq)){
      myvar = colnames(temp)[myseq[v]] 
      out[i,2*v+2] = median (temp[,myvar], na.rm=T)
      out[i,2*v+3] = sd (temp[,myvar], na.rm=T)
    }
    
    if (verbose) {
      time_elapsed = as.numeric ((proc.time() - ptm)[3])
      speed = time_elapsed / sum(filesizes[1:i])
      time_left = speed * sum(filesizes[min((i+1),length(filesizes)):length(filesizes)])
      minutes = time_left %/% 60; seconds = round (time_left %% 60, 0)
      if (minutes>0) {
        report = paste0('Done with file ',i, ' of ', length(filenames), '. Estimated time left: ', minutes, ' min ', seconds, ' sec.')
      } else {
        report = paste0('Done with file ',i, ' of ', length(filenames), '. Estimated time left: ', seconds, ' sec.')
      }     
      if (i<length(filenames)){
        print(report)
      } else {
        minutes = time_elapsed %/% 60; seconds = round (time_elapsed %% 60, 0)
        print (paste0('Processed ', length(filenames), ' files in ', minutes, ' min ', seconds,' sec.'))
      }
    }
  }
  return (out)
}

# example of use: 
# myfolder = '/home/allgoodguys/Documents/Studying/Lund_PhD/epistles/04_article_nat-actor/analysis/plots_ut590_pitch_july2016'
# df = analyzeFolder (myfolder, savePlot=F, plot=F, verbose=T, savePlotPath = paste0(myfolder,'/'))
write.csv(df, '~/Downloads/ut590_july.csv')

