setwd('/home/allgoodguys/Documents/Studying/Lund_PhD/epistles/04_article_nat-actor/analysis')

library (randomForest)

scaleWithinCorpus = c(T,F)[2] # if false, we scale the whole corpus (basically analogous to not scaling at all, since RF just finds cutoff values). If true, we scale real/fake stuff separately (good for illustrating the effect of extreme values on classification accuracy, eg real-fake transfer improves for fear and anger)

all = read.csv ('all_july2016.csv')
all = droplevels (subset(all, all$emotion!='joy' & all$emotion!='effort' & all$emotion!='achievement'))
all$corpus = as.factor ( ifelse (all$corpus=='ut','YouTube','Other') )

rescale_df = function(df){
  df$pitch_manually = ifelse (!is.na(df$pitch_manually), log2(df$pitch_manually), NA)
  df$pitch_sd = log2(df$pitch_sd+1) # b/c it's occasionally 0
  df$quartile25_median = log2(df$quartile25_median+1)
  df$quartile25_sd = log2(df$quartile25_sd+1)
  return (df)
}

all = rescale_df (all)

vars = c("pitch_manually", "pitch_sd", "HNR_median","HNR_sd", "quartile25_median", "quartile25_sd", "harmonics_median",  "specSlope_median", "specSlope_sd", "syllableLength_median", "syllableLength_sd", "interBurst_median", "interBurst_sd", "ampl_voiced_median", "ampl_voiced_sd", "voiced")
df = all[, c("corpus","emotion",vars)]

if (scaleWithinCorpus==F) { # just scale the whole dataset (789 sounds)
  df[,3:ncol(df)] = scale (df[,3:ncol(df)])
  df[is.na(df)] <- 0
} 

for (c in levels(df$corpus)){
  assign(c, droplevels (subset(df, df$corpus==c)))
} # now we have 2 (or 8) separate dataframes (with NA's replaced by zeroes - NB! After scaling)


train <- YouTube; test <- Other 
# train <- Other; test <- YouTube

if (scaleWithinCorpus){ # scale separately for training and testing set
  train[,3:ncol(train)] = scale (train[,3:ncol(train)])
  train[is.na(train)] <- 0
  test[,3:ncol(test)] = scale (test[,3:ncol(test)])
  test[is.na(test)] <- 0
}

# loop starts here
noIter = 1000
ptm <- proc.time()  
output = array (dim=c(8, 2, noIter))

for (i in 1:noIter) {
  model.boot = randomForest (emotion ~ harmonics_median + specSlope_median+specSlope_sd + quartile25_median+quartile25_sd + pitch_manually+pitch_sd + HNR_median+HNR_sd + ampl_voiced_median+ampl_voiced_sd + syllableLength_median+syllableLength_sd + interBurst_median+interBurst_sd +voiced, train, importance=TRUE, ntree=1000, strata=train$emotion, sampsize=rep(min(table(train$emotion)), length(unique(train$emotion))))   
  
  output [,1,i] = 1-apply (model.boot$err.rate, 2, mean)  # Accuracies for train-train (internal cross-validation). Order: overall, amus, anger...sadness 
  test$pred = as.character (predict (model.boot,test))  # confusion: table(test$emotion,test$pred)
  output [,2,i] = c( mean (test$pred==test$emotion), tapply(test$pred==test$emotion, test$emotion, FUN = mean) )
  
  time_elapsed = as.numeric ((proc.time() - ptm)[3])
  time_left = time_elapsed/i * (noIter-i)
  minutes = time_left %/% 60; seconds = round (time_left %% 60, 0)
  if (minutes>0) {
    report = paste0('Done with iteration ',i, ' of ', noIter, '. Time left: ', minutes, ' min ', seconds, ' sec.')
  } else {
    report = paste0('Done with iteration ',i, ' of ', noIter, '. Time left: ', seconds, ' sec.')
  }
  print (report)
}

# result = round ( apply (output, c(1,2), mean)*100, 0)
result = apply (output, c(1,2), function(x){
  paste0(round(mean(x)*100, 0), ' [', round(quantile(x, probs=.025)*100, 0), ', ', round(quantile(x, probs=.975)*100, 0), ']')
})
rownames(result) = c('overall', levels(train$emotion))
colnames(result) = c('train','test')
result = result[c(2:8, 1), ]

write.csv (result, '~/Downloads/temp.csv')




### Training and testing on pooled data (both authentic and acted)
# loop starts here
noIter = 1000
ptm <- proc.time()  
output = array (dim=c(8, 2, noIter))

for (i in 1:noIter) {
  perSound = data.frame(sound=1:nrow(df), corpus=df$corpus, emotion=df$emotion, predicted_em=NA)
  model.boot = randomForest (emotion ~ harmonics_median + specSlope_median+specSlope_sd + quartile25_median+quartile25_sd + pitch_manually+pitch_sd + HNR_median+HNR_sd + ampl_voiced_median+ampl_voiced_sd + syllableLength_median+syllableLength_sd + interBurst_median+interBurst_sd +voiced, df, importance=TRUE, ntree=1000, strata=df$emotion, sampsize=rep(min(table(df$emotion)), length(unique(df$emotion))))   
  
  perSound$predicted_em = model.boot$predicted
  ut = perSound[perSound$corpus=='YouTube',]
  other = perSound[perSound$corpus=='Other',]
  output [1,1,i] = as.numeric ( aggregate ((predicted_em==emotion)~1, ut, mean) )
  output [ (2:8) ,1,i] = aggregate ((predicted_em==emotion)~emotion, ut, mean)[,2] # accuracies for YouTube
  output [1,2,i] = as.numeric ( aggregate ((predicted_em==emotion)~1, other, mean) )
  output [ (2:8) ,2,i] = aggregate ((predicted_em==emotion)~emotion, other, mean)[,2] # accuracies for Other
  
  time_elapsed = as.numeric ((proc.time() - ptm)[3])
  time_left = time_elapsed/i * (noIter-i)
  minutes = time_left %/% 60; seconds = round (time_left %% 60, 0)
  if (minutes>0) {
    report = paste0('Done with iteration ',i, ' of ', noIter, '. Time left: ', minutes, ' min ', seconds, ' sec.')
  } else {
    report = paste0('Done with iteration ',i, ' of ', noIter, '. Time left: ', seconds, ' sec.')
  }
  print (report)
}

result = apply (output, c(1,2), function(x){
  paste0(round(mean(x)*100, 0), ' [', round(quantile(x, probs=.025)*100, 0), ', ', round(quantile(x, probs=.975)*100, 0), ']')
})
rownames(result) = c('overall', levels(df$emotion))
colnames(result) = c('YouTube','Other')
result = result[c(2:8, 1), ]
write.csv (result, '~/Downloads/temp.csv')




### ### Training and testing on HALF the pooled data (50% of authentic and 50% of acted) # or just 181 of each (2*181 = 362 = the total size of acted)

# loop starts here
noIter = 1000
ptm <- proc.time()  
output = array (dim=c(8, 2, noIter))

for (i in 1:noIter) {
  # stratified random sampling by corpus and emotion
  train_ut = lapply(split(YouTube, YouTube$emotion), function(x) x[sample(1:nrow(x), floor(nrow(x)/2.3), replace=FALSE),])
  train_ut = do.call (rbind, train_ut)
  train_other = lapply(split(Other, Other$emotion), function(x) x[sample(1:nrow(x), nrow(x)/2, replace=FALSE),])
  train_other = do.call (rbind, train_other)
  train <- rbind (train_ut, train_other)
  
  mynames = sapply (rownames(train), function(x) unlist (strsplit (as.character(x), split="[.]"))[2]) # numbers of files included in the training set
  test <- df [!rownames(df) %in% mynames, ] # non-sampled files make up the testing set
  
  perSound = data.frame(sound=1:nrow(test), corpus=test$corpus, emotion=test$emotion, predicted_em=NA)
  model.boot = randomForest (emotion ~ harmonics_median + specSlope_median+specSlope_sd + quartile25_median+quartile25_sd + pitch_manually+pitch_sd + HNR_median+HNR_sd + ampl_voiced_median+ampl_voiced_sd + syllableLength_median+syllableLength_sd + interBurst_median+interBurst_sd +voiced, train, importance=F, ntree=1000, strata=train$emotion, sampsize=rep(min(table(train$emotion)), length(unique(train$emotion))))   
  test$pred = as.character (predict (model.boot,test))  # confusion: table(test$emotion,test$pred)
  global = aggregate ( (test$pred==test$emotion)~corpus, test, mean)
  per_em = aggregate ( (test$pred==test$emotion)~emotion+corpus, test, mean)
  
  output [,1,i] = c ( per_em[8:14,3], global[2,2] ) # YouTube
  output [,2,i] = c ( per_em[1:7,3], global[1,2] ) # Other
  
  time_elapsed = as.numeric ((proc.time() - ptm)[3])
  time_left = time_elapsed/i * (noIter-i)
  minutes = time_left %/% 60; seconds = round (time_left %% 60, 0)
  if (minutes>0) {
    report = paste0('Done with iteration ',i, ' of ', noIter, '. Time left: ', minutes, ' min ', seconds, ' sec.')
  } else {
    report = paste0('Done with iteration ',i, ' of ', noIter, '. Time left: ', seconds, ' sec.')
  }
  print (report)
}

result = apply (output, c(1,2), function(x){
  paste0(round(mean(x)*100, 0), ' [', round(quantile(x, probs=.025)*100, 0), ', ', round(quantile(x, probs=.975)*100, 0), ']')
})
rownames(result) = c(levels(df$emotion), 'overall')
colnames(result) = c('YouTube','Other')
result # table(test$corpus); table(train$corpus)
# write.csv (result, '~/Downloads/temp.csv')
