### Script for finding syllables and bursts of acoustic energy in .wav files
# NB: default settings are the result of slow and painstaking optimization against manual segmentation. Feel free to modify!

## other things to try, ideas: 
# - run the function iteratively: use the number and spacing of bursts detected at first pass to fine-tune the parameters (smoothing, threshold for syllables, etc) and do a second pass 
# - try to define dynamic thresholds for syllable detection (how?)

library(tuneR)
library(seewave)

findSyllablesAndBursts = function(soundfile, timemin=NULL, timeminIdx = 118, thresholdIdx = 89, smooth_points = 1200, smooth_overlap = 90, heightIdx = 320, shortestSyllable=39, shortestBreakBwSyl=50, peak_above_global_max_Idx = 12,  plot=F, savePlot=F, savePlotPath='/home/allgoodguys/Downloads/tempPlots/', verbose=F, merge_syllables_with_short_intervals_in_between=T) {

  # smooth_points: length of smoothing window, in points (so depends on the sampling rate - the default is for 44100 but it is adjusted automatically for other sampling rates, so for eg 22050 the effective lenght will be half that given, etc)
  # smooth_overlap: overlap between smoothing windows (%) - the higher, the more accurate, but also slower
  # thresholdIdx: amplitude threshold for syllable detection (% of global mean amplitude of smoothed envelope)
  # shortestSyllable: minimum acceptable length of syllables (ms)
  # shortestBreakBwSyl: minimum acceptable break between syllables (ms). Syllables separated by less time are merged (works only if merge_syllables_with_short_intervals_in_between=T)
  # timemin is the minimum time between two consecutive cycles (ms). NB: if specified explicitly, it overrides the timeminIdx setting! 
  # timeminIdx (%) is the multiplier of the default timemin value (median syllable length or, if no syllables are detected, same number as shortestSyllable). Only used if timemin is not specified. Large number improves detection of unusually broad shallow peaks, smaller number improves the detection of sharp narrow peaks
  # heightIdx specifies the factor by which a burst has to exceed the local minimum on the LEFT over time interval [time_burst-timemin/2 : time_burst]
  # NB: tested on file > 270 ms in duration. May break down on shorter files, depending on the parameter settings! 
  
  # open, pre-format and low-pass filter the sound file by thinning
  sound = readWave (soundfile)
  samplingRate = sound@samp.rate
  sound = sound@left
  if (min(sound)>0) {sound = scale(sound)}
  sound = sound / max( abs(max(sound)), abs(min(sound)))
  smooth_points = smooth_points * samplingRate / 44100
  df = env (sound, f=samplingRate, msmooth=c(smooth_points, smooth_overlap), fftw=T, plot=F)
  timestep = 1000 / samplingRate * (length(sound)/nrow(df)) # time step in the envelope, ms
  df = data.frame (ma = df, time = (1:nrow(df)*timestep))
  threshold = thresholdIdx/100 * mean(df$ma) 
  
  ## SYLLABLES
  # find syllables and get descriptives
  df$aboveThres = ifelse (df$ma>threshold, 1, 0)
  temp = data.frame (value = rle(df$aboveThres)[[2]], count = rle(df$aboveThres)[[1]])  # find strings of TTTTT
  temp$idx = 1:nrow(temp) # a convoluted way of tracing the time stamp in the output of rle
  tempShort = na.omit ( subset (temp, temp$value==1 & temp$count>ceiling(shortestSyllable/timestep)) )  # exclude segments of length < shortestSyllable or below threshold 
  
  noSyllables = nrow (tempShort)
  syllableLength_mean = mean (tempShort$count) * timestep  # to convert from steps to ms
  syllableLength_sd = sd (tempShort$count) * timestep
  
  # save the time of each syllable for plotting
  syllable=numeric(); timeStart=numeric(); timeEnd=numeric()
  if (noSyllables>0) {
    for (s in 1:noSyllables) {
      syllable[s] = s
      syllable_index_in_temp = tempShort$idx[s]
      if (s==1 & temp$value[1]==1) { # if the sounds begins with a syllable
        timeStart[s] = 0  # the first syllable begins at zero
      } else {  # if the sound begins with ~silence
        timeStart[s] = sum(temp$count[ 1 : (syllable_index_in_temp-1) ]) * timestep
      }
      timeEnd[s] = timeStart[s] + tempShort$count[s] * timestep
    }
  } 
  syllables = data.frame ('syllable'=syllable, 'timeStart'=timeStart, 'timeEnd'=timeEnd)
  
  # Optional: merge syllables with very short intervals in between them
  if (merge_syllables_with_short_intervals_in_between) {
    syllables = mergeSyllables (syllables, shortestBreakBwSyl=shortestBreakBwSyl)
    syllables$syllable = 1:nrow(syllables)
  }
  
  ## BURSTS
  # calculate the window for analyzing bursts based on the median duration of syllables (if no syllables are detected, just use the specified shortest acceptable syllable length)
  if (is.null(timemin)) {
    aa = round (median (tempShort$count) * timeminIdx * timestep/100, 0)  # or syllableLength_mean
    if (is.na(aa)) {timemin=as.numeric(shortestSyllable)} else {timemin=aa}  # we use "try" in case tempShort$count is not defined
    # print (timemin)
  }
  n = floor (timemin/timestep)
  
  # set vectors to store info
  noBursts = 0
  number=numeric() 
  time=numeric() 
  ampl=numeric() 
  
  for (i in 1:nrow(df)) {
    
    # for each datapoint, compare it with the local minima to the left/right over ± timemin ms
    if (i>n) {
      local_min_left = min (df$ma [(i-n):i])
      dif_left = df$ma[i]-local_min_left
    } else {
      local_min_left = 0
      dif_left=df$ma[i]
      }             # close to the beg of the file, local_min_left=0
    if (i<(nrow(df)-n-1)) {
      local_min_right = min (df$ma [i:(i+n)]) # lowest ampl over timemin ms on the right
      dif_right = df$ma[i] - local_min_right
    } else {
      local_min_right = 0 
      dif_right=df$ma[i] # just in case we want to evaluate both sides of a peak
    }
    
    # define the window for analysis (differs from ± timemin because we have to consider the beginning and end of file)
    if (i>n) {limleft=i-n} else {limleft=1}
    if (i<(nrow(df)-n-1)) {limright=i+n} else {limright = nrow(df)}
    
    # DEFINITION OF A BURST FOLLOWS!!!
    if (
      df$ma[i] == max(df$ma[limleft:limright]) &  # it is a local maximum over ± timemin
      df$ma[i]/local_min_left > heightIdx/100 &   # it exceeds the local minimum on the LEFT by a factor of heightIdx
      df$ma[i] > peak_above_global_max_Idx / 100 * max(df$ma) # the peak is above a certain % of the global maximum
    ) {
      noBursts=noBursts+1    # keep track of the total number of bursts, their time and amplitude
      number [noBursts] = noBursts 
      time [noBursts] = i * timestep
      ampl [noBursts] = df$ma [i] 
      if (verbose) {writeLines(paste("time of burst no.", noBursts, " =", time [noBursts]))}
    }    
  } 
  
  # get descriptives for bursts
  bursts = data.frame(burst=number, time=time, ampl=ampl)
  interval=numeric()
  noBursts = nrow(bursts)
  
  # get the mean interval between bursts and its sd
  if (noBursts>1) {
    interval [1] = NA
    for (i in 2:noBursts) {
      interval [i] = bursts$time[i] - bursts$time [i-1]
    }
  } else {interval=rep(NA, noBursts)}
  mean_interval = mean (interval, na.rm=TRUE)
  median_interval = median (interval, na.rm=TRUE)
  sd_interval = sd (interval, na.rm=TRUE)
  
  # plotting and saving plots (optional)
  if (plot) {
    plotname = tail (unlist (strsplit(soundfile, '/')),n=1)
    plotname = substring (plotname, 1, nchar(plotname)-4)
    plot ( df$time, df$ma, type='l', col='green', xlab='Time, ms', ylab='Amplitude', main=plotname)
    points (time,ampl, col='red', cex=3, pch=8) 
    for (s in 1:nrow(syllables)) {
      segments (x0=syllables$timeStart[s], y0=threshold, x1=syllables$timeEnd[s], y1=threshold, lwd=2, col='blue')
    }
  }
  
  if (savePlot) {
    plotname = tail (unlist (strsplit(soundfile, '/')),n=1)
    plotname = substring (plotname, 1, nchar(plotname)-4)
    jpeg(file = paste0 (savePlotPath, plotname, ".jpg"), 900,500)
    plot ( df$time, df$ma, type='l', col='green', xlab='Time, ms', ylab='Amplitude')
    points (time,ampl, col='red', cex=3, pch=8) 
    for (s in 1:nrow(syllables)) {
      segments (x0=syllables$timeStart[s], y0=threshold, x1=syllables$timeEnd[s], y1=threshold, lwd=2, col='blue')
    }
    dev.off()
  } 
  
  # return a dataframe containing descriptives for syllables and bursts
  result = data.frame (noSyllables=nrow(syllables), syllableLength_mean=mean(syllables$timeEnd-syllables$timeStart), syllableLength_median=median(syllables$timeEnd-syllables$timeStart), syllableLength_sd=sd(syllables$timeEnd-syllables$timeStart), noBursts=noBursts, interBurst_mean=mean_interval, interBurst_median = median_interval, interBurst_sd=sd_interval)
  return (result)
}


findSyllablesAndBurstsInFolder = function (myfolder, timemin=NULL, timeminIdx = 118, thresholdIdx = 89, smooth_points = 1200, smooth_overlap = 90, heightIdx = 320, shortestSyllable=39, shortestBreakBwSyl=50, peak_above_global_max_Idx = 12, plot=F, savePlot=F, savePlotPath='/home/allgoodguys/Downloads/tempPlots/', verbose=F, reportTime=F, merge_syllables_with_short_intervals_in_between=T) {  # calls findSyllablesAndBursts to analyze a whole folder rather than one sound at a time
  
  ptm <- proc.time()   # timing
  filenames = list.files(myfolder, pattern="*.wav", full.names=TRUE)  # open all .wav files in folder
  output = matrix(ncol=9,nrow=length(filenames))
  
  for (i in 1:length(filenames)) {
    output[i,1] = tail (unlist (strsplit(filenames[i], '/')),n=1)
    output[i, 2:9] = as.matrix (findSyllablesAndBursts(filenames[i], timemin=timemin, timeminIdx = timeminIdx, thresholdIdx = thresholdIdx, smooth_points = smooth_points, smooth_overlap = smooth_overlap, heightIdx = heightIdx, shortestSyllable=shortestSyllable, shortestBreakBwSyl=shortestBreakBwSyl, peak_above_global_max_Idx, plot=plot, savePlot=savePlot, savePlotPath=savePlotPath, verbose=verbose, merge_syllables_with_short_intervals_in_between=merge_syllables_with_short_intervals_in_between))
  }
  colnames(output) = c('sound','noSyllables','syllableLength_mean','syllableLength_median','syllableLength_sd','noBursts','interBurst_mean','interBurst_median','interBurst_sd')
  
  if (reportTime) {print ( paste (length(filenames), 'audio files processed in', round (as.numeric ((proc.time() - ptm)[3]),0), 's.') )}
  return (as.data.frame(output))
}

## internal helper functions 

remove_NA = function (df) { # removes NA's from the output of findSyllablesAndBurstsInFolder replacing them with an arbitrarily chosen large number, say 20000. 
  
  df$syllableLength_mean = as.character (df$syllableLength_mean)
  df$syllableLength_median = as.character (df$syllableLength_median)
  df$syllableLength_sd = as.character (df$syllableLength_sd)
  df$interBurst_mean = as.character (df$interBurst_mean)
  df$interBurst_median = as.character (df$interBurst_median)
  df$interBurst_sd = as.character (df$interBurst_sd)
  for (r in 1:nrow(df)) {
    if (is.na(df$syllableLength_mean[r]) | df$syllableLength_mean[r]=='NaN') {df$syllableLength_mean[r]=20000}  
    if (is.na(df$syllableLength_median[r]) | df$syllableLength_median[r]=='NaN') {df$syllableLength_median[r]=20000} 
    if (is.na(df$syllableLength_sd[r]) | df$syllableLength_sd[r]=='NaN') {df$syllableLength_sd[r]=20000}  
    if (is.na(df$interBurst_mean[r]) | df$interBurst_mean[r]=='NaN') {df$interBurst_mean[r]=20000}  
    if (is.na(df$interBurst_median[r]) | df$interBurst_median[r]=='NaN') {df$interBurst_median[r]=20000}  
    if (is.na(df$interBurst_sd[r]) | df$interBurst_sd[r]=='NaN') {df$interBurst_sd[r]=20000} 
  }
  df$syllableLength_mean = as.numeric (df$syllableLength_mean)
  df$syllableLength_median = as.numeric (df$syllableLength_median)
  df$syllableLength_sd = as.numeric (df$syllableLength_sd)
  df$interBurst_mean = as.numeric (df$interBurst_mean)
  df$interBurst_median = as.numeric (df$interBurst_median)
  df$interBurst_sd = as.numeric (df$interBurst_sd)
  
  return (df)
}

mergeSyllables = function (df, shortestBreakBwSyl=80) {  # merges syllables if they are separated by < shortestBreakBwSyl ms
  temp = df  # df is a dataframe listing syllables with timeStart and timeEnd
  i=1
  while (i<nrow(temp)) {
    while (temp$timeStart[i+1]-temp$timeEnd[i] < shortestBreakBwSyl) {
      temp$timeEnd[i] = temp$timeEnd[i+1]
      if ((i+1)<nrow(temp)) {temp = temp[-(i+1),]} else {break}
    }
    i=i+1
  }
  return (temp)
}

plot_some_examples = F  # check manually on some problematic files from different categories:
if (plot_some_examples) {
  path = '/home/allgoodguys/Documents/Studying/Lund_cognitive-science/00_master/cartoons_clips/all_shrunk_260/'
  files = c('ut_exams_05-f-laugh', 'ut_fear-bungee_11', 'ut_fear_23-scream-laugh', 'ut_fear_29-f-scream', 'ut_fear_49-m-scream', 'ut_fear_51-m-scream', 'ut_laugh-amused_05-f', 'ut_laugh-amused_31-m', 'ut_laugh-amused_53-f', 'ut_pain_47-birth', 'ut_sad-cry_16-m', 'ut_sad-cry_38-f')
  length(files)
  
  layout (matrix(nrow=3, ncol=4, 1:12, byrow=T))
  for (f in sort(files)) {
    myfile = paste0(path,f,'.wav')
    findSyllablesAndBursts (myfile, plot=T)
  }
  layout (matrix(c(1,1)))
}

## Example of use:
# myfolder = '/home/allgoodguys/Documents/Studying/Lund_PhD/epistles/04_article_nat-actor/analysis/ut_590_numbered'
myfolder = '/home/allgoodguys/Documents/Studying/Lund_PhD/epistles/04_article_nat-actor/analysis/simon-thomas'
df_syl = findSyllablesAndBurstsInFolder(myfolder, reportTime=T)
write.csv (df_syl, '~/Downloads/temp.csv')
# df = remove_NA (df_syl)
# write.csv(df, '~/Downloads/bursts_no-na.csv')

findSyllablesAndBursts (soundfile='~/Downloads/simon/simon_amusement_a041b.wav', timemin=NULL, timeminIdx = 118, thresholdIdx = 89, smooth_points = 1200, smooth_overlap = 90, heightIdx = 320, shortestSyllable=39, shortestBreakBwSyl=50, peak_above_global_max_Idx = 12,  plot=T, savePlot=F, savePlotPath='/home/allgoodguys/Downloads/tempPlots/', verbose=F, merge_syllables_with_short_intervals_in_between=T)

findSyllablesAndBursts (soundfile='~/Downloads/simon/simon_amusement_a041b.wav', plot=T) 